/*

    PortoSeguro - Light Application Launcher
    Copyright (C) 2000 Felipe Paulo Guazzi Bergo
    bergo@seul.org

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <gtk/gtk.h>
#include "dock.h"
#include "pixmap.h"
#include "common.h"
#include "about.h"

#include "square.xbm"
#include "graydock.xpm"
#include "sea.xpm"

int icon_path_count=0;
char icon_paths[50][256];
char *icon_path_p[50];

GdkBitmap *mask;
GdkBitmap *squaremask=NULL;

GList *offspring=NULL;

int narray[36];

int halign,valign,direc;

int speed=11; /* 2,3,6,11,22 - 66 must be divisible by speed*/
int compress=1;
int bcount=1;
struct PortoSeguroDock bdock[36]; /* you _don't_ have enough screen :-) ! */

struct HueMask activehue;

int main(int argc,char **argv) {
  int i;

  gtk_init(&argc,&argv);  
  gdk_rgb_init();

  for(i=0;i<36;i++)
    narray[i]=i;
  for(i=0;i<50;i++)
    icon_path_p[i]=&(icon_paths[i][0]);

  read_dock();
  init_dock();
  
  gtk_idle_add(check_children,NULL);
  gtk_timeout_add(100,startup,NULL);
  gtk_main();
  
  gdk_bitmap_unref(squaremask);
  
  for(i=0;i<bcount;i++) {
    gtk_widget_destroy(bdock[i].square);
    g_free(bdock[i].gray);
    g_free(bdock[i].purple);
  }

  return 0;
}

void init_dock(void)
{
  GtkWidget *b1,*da;
  GtkStyle *s;
  int i,bx,by,w,h,iby;

  w=gdk_screen_width();
  h=gdk_screen_height();

  bx=w-66;
  by=h-65;

  direc=-1;

  if (!halign)
    bx=2;
  if (!valign) {
    by=1;
    direc=1;
  }

  iby=by;

  for(i=0;i<bcount;i++) {
    b1=gtk_window_new(GTK_WINDOW_POPUP);
    gtk_window_set_wmclass(GTK_WINDOW(b1),"PortoSeguro","DockButton");
    gtk_window_set_title(GTK_WINDOW(b1),"OMNI");
    bdock[i].square=b1;
    bdock[i].x=bx;
    bdock[i].pos=i;
    gtk_widget_set_uposition(b1,bx,
			     bdock[i].y=
			     ((compress)?iby:(by+(direc*i*66))));
    gtk_window_set_default_size(GTK_WINDOW(b1),64,64);
    gtk_widget_realize(b1);
    
    da=gtk_drawing_area_new();
    gtk_drawing_area_size(GTK_DRAWING_AREA(da),64,64);
    gtk_widget_set_events(da,GDK_EXPOSURE_MASK|GDK_BUTTON_PRESS_MASK);
    gtk_container_add(GTK_CONTAINER(b1),da);
    
    s=gtk_widget_get_style(b1);
    
    if (squaremask==NULL) {
      squaremask=gdk_bitmap_create_from_data(b1->window,
					     square_bits,
					     64,
					     64);
    }
    gtk_widget_shape_combine_mask(b1,squaremask,0,0);

    if (i) {
      if (!bdock[i].applet) {
	bdock[i].gray=make_rgb_mix_file(bdock[i].icon,graydock_xpm,NULL);
	bdock[i].purple=make_rgb_mix_file(bdock[i].icon,graydock_xpm,
					  &activehue);
      } else {
	bdock[i].gray=make_rgb_mix_file("anyone",graydock_xpm,NULL);
	bdock[i].purple=make_rgb_mix_file("anyone",graydock_xpm,&activehue);
      }
    } else {
      bdock[0].gray=make_rgb_mix_data(sea_xpm,graydock_xpm,NULL);
      bdock[0].purple=make_rgb_mix_data(sea_xpm,graydock_xpm,&activehue);
    }

    gtk_signal_connect(GTK_OBJECT(da),"expose_event",
		       GTK_SIGNAL_FUNC(expose),narray+i);
    gtk_signal_connect(GTK_OBJECT(b1),"delete_event",
		       GTK_SIGNAL_FUNC(die),NULL);
    gtk_signal_connect(GTK_OBJECT(b1),"destroy",
		       GTK_SIGNAL_FUNC(dead),NULL);
    gtk_signal_connect(GTK_OBJECT(b1),"button_press_event",
		       GTK_SIGNAL_FUNC(click),narray+i);
    gtk_widget_show(da);
    gtk_widget_show(b1);
  }

  /* correct order */
  for(i=bcount-1;i>=0;i--)
    gdk_window_raise(bdock[i].square->window);
}

gboolean expose(GtkWidget *a,GdkEventExpose *b,gpointer c)
{
  int v;
  char *buf;

  v=*(int *)(c);

  /* paint colored if (hold & running) */
  if ((bdock[v].running)&&(bdock[v].hold)) {
      gdk_draw_rgb_image(a->window,
			 a->style->fg_gc[GTK_WIDGET_STATE(a)],
			 0, 0,
			 64, 64,
			 GDK_RGB_DITHER_NORMAL,
			 bdock[v].purple,
			 3*64);
      return FALSE;
  }
  /* paint gray if (hold & !running) | (!justran) */
  if ((!bdock[v].justran) || (bdock[v].hold)) {
    gdk_draw_rgb_image(a->window,
		       a->style->fg_gc[GTK_WIDGET_STATE(a)],
		       0, 0, 64, 64,
		       GDK_RGB_DITHER_NORMAL,
		       bdock[v].gray, 3*64);
    return FALSE;
  }
  /* decaying highlight if (justran & !hold) */
  if (bdock[v].justran>20) {
    bdock[v].justran=8;
    gtk_timeout_add(40,blink_callback,(gpointer)(&bdock[v]));
  } else {
    buf=rgb_morph(bdock[v].gray,bdock[v].purple,
		  0.20*(double)(bdock[v].justran));
    gdk_draw_rgb_image(a->window,
		       a->style->fg_gc[GTK_WIDGET_STATE(a)],
		       0, 0,64, 64,GDK_RGB_DITHER_NORMAL,
		       buf,3*64);
    g_free(buf);
  }
  return FALSE;
}

gboolean die(GtkWidget *w,GdkEvent *e,gpointer data)
{
  return FALSE;
}

gboolean dead(GtkWidget *w,gpointer data)
{
  return FALSE;
}

void read_dock(void)
{
  FILE *f;
  char *p;
  char tmp[512],aux[256];
  int i;

  icon_path_count=0;
  activehue.r = 0.6667;
  activehue.g = 0.6667;
  activehue.b = 1.0000;
  halign=valign=1;

  strcpy(bdock[0].icon,"null");
  strcpy(bdock[0].exec,"portoseguro.conf");
  bdock[0].hold=1;
  bdock[0].applet=0;
  bdock[0].running=0;
  bdock[0].justran=0;

  p=getenv("HOME");
  sprintf(tmp,"%s/.sapphire/PortoSeguro",p);
  f=fopen(tmp,"r");

  if (!f) {
    sprintf(tmp,"%s/.PortoSeguro",p);
    f=fopen(tmp,"r");
    if (!f)
      return;
  }

  while(fgets(tmp,511,f)) {
    if (strlen(tmp)<6)
      continue;
    if (strstr(tmp,"#Item#")) {
      fgets(bdock[bcount].icon,255,f);
      strip_n(bdock[bcount].icon);
      fgets(bdock[bcount].exec,511,f);
      strip_n(bdock[bcount].exec);
      fgets(tmp,511,f);
      strip_n(tmp);
      bdock[bcount].hold=atoi(tmp);
      bdock[bcount].running=0;
      bdock[bcount].justran=0;
      fgets(tmp,511,f);
      strip_n(tmp);
      bdock[bcount].applet=atoi(tmp);
      bcount++;
    }
    if (strstr(tmp,"#ActivatedColor#")) {
      fgets(aux,255,f);
      activehue.r=atof(strtok(aux," \t\n"));
      activehue.g=atof(strtok(NULL," \t\n"));
      activehue.b=atof(strtok(NULL," \t\n"));
    }
    if (strstr(tmp,"#Place#")) {
      fgets(aux,255,f);
      halign=atoi(strtok(aux," \t\n"));
      valign=atoi(strtok(NULL," \t\n"));
    }
    if (strstr(tmp,"#AnimeSpeed#")) {
      fgets(aux,255,f);
      speed=atoi(strtok(aux," \t\n"));
    }
    if (strstr(tmp,"#IconPath#")) {
      icon_path_count=0;
      while(1) {
	fgets(icon_paths[icon_path_count],255,f);
	if (strstr(icon_paths[icon_path_count],"#IconPath#"))
	  break;
	strip_n(icon_paths[icon_path_count++]);
      }
    }
  }
  fclose(f);

  if (!icon_path_count) {
    icon_path_count=2;
    strcpy(icon_paths[0],"/usr/share/icons");
    strcpy(icon_paths[1],"/usr/local/share/icons");
  }

  /* prepend path if necessary - it should not be */
  for(i=0;i<bcount;i++)
    assure_icon(bdock[i].icon,icon_path_p,icon_path_count);
}

gboolean click(GtkWidget *a,GdkEventButton *b,gpointer c)
{
  int v;
  v=*(int *)c;

  if (compress)
    v=0;
  
  if ((!v)&&(b)) { /* middle click on setup opens about */
    switch(b->button) {
    case 2:
      open_about();
      return FALSE;
    case 3:
      compress=!compress;
      if (compress)
	fx_compress();
      else
	fx_uncompress();
      return FALSE;
    }
  }

  if ((bdock[v].running)&&(bdock[v].hold))
    return FALSE;
  if (bdock[v].applet)
    return FALSE;

  bdock[v].running=1;
  bdock[v].justran=50;
  bdock[v].pid=fork();
  
  if (!bdock[v].pid) {
    execlp("/bin/sh","sh","-c",bdock[v].exec,NULL);
    bdock[v].running=0;
  }

  gtk_widget_queue_resize(bdock[v].square);
  return FALSE;
}

gboolean check_children(gpointer data)
{
  int i;
  char *p,tmp[256];
  
  for(i=0;i<bcount;i++) {
    if (bdock[i].running)
      if (waitpid(bdock[i].pid,NULL,WNOHANG)==bdock[i].pid) {
	bdock[i].running=0;
	gtk_widget_queue_resize(bdock[i].square);
      }
  }

  /* got zombies still ? */
  waitpid(-1,NULL,WNOHANG);

  p=getenv("HOME");
  sprintf(tmp,"%s/.PortoSeguroDock.update.yourself",p);
  if (file_exists(tmp)) {
    unlink(tmp);
    respawn_dock();
  }

  usleep(100000);
  return TRUE;
}

void respawn_dock(void)
{
  int i;
  struct LostChild *lc;
  GList *pt;

  /* nicy effect */
  if (!compress) {
    compress=1;
    fx_compress();
  }

  /* build offspring list */

  for(i=0;i<bcount;i++)
    if (bdock[i].running) {
      if (bdock[i].applet) {
	kill(bdock[i].pid,SIGTERM);
	continue;
      }
      lc=(struct LostChild *)g_malloc0(sizeof(struct LostChild));
      lc->pid=bdock[i].pid;
      strcpy(lc->exec,bdock[i].exec);
      offspring=g_list_append(offspring,(gpointer)lc);
    }

  for(i=0;i<bcount;i++) {
    gtk_widget_destroy(bdock[i].square);
    g_free(bdock[i].gray);
    g_free(bdock[i].purple);
  }

  gdk_bitmap_unref(squaremask);

  bcount=1;
  squaremask=NULL;

  read_dock();
  
  /* adjust to previous state */

  for(pt=offspring;pt!=NULL;pt=g_list_next(pt)) {
    lc=(gpointer)(pt->data);
    if (waitpid(lc->pid,NULL,WNOHANG)==0) {
      for(i=0;i<bcount;i++)
	if (!strcmp(bdock[i].exec,lc->exec)) {
	  bdock[i].running=1;
	  bdock[i].pid=lc->pid;
	  break;
	}
    }
    g_free((gpointer)lc);
  }

  g_list_free(offspring);
  offspring=NULL;

  /* respawn dock */

  init_dock();

  while(gtk_events_pending())
    gtk_main_iteration();
  gdk_flush();

  usleep(5000);

  compress=0;
  fx_uncompress();
}

gint blink_callback(gpointer data)
{
  struct PortoSeguroDock *p;
  p=(struct PortoSeguroDock *)data;
  gtk_widget_queue_resize(p->square);
  return(!!(--(p->justran)));
}

void fx_compress(void)
{
  int i,c;
  int py[36],dy;

  for(i=0;i<bcount;i++)
    py[i]=bdock[i].y;

  if (66%speed)
    speed=11;

  dy=1;
  if (valign)
    dy=gdk_screen_height()-65;

  while(1) {
    c=0;
    for(i=0;i<bcount;i++) {
      if (bdock[i].y != dy) {
	++c;
	bdock[i].y-=direc*speed;
	gdk_window_move(bdock[i].square->window,bdock[i].x,bdock[i].y);
      }
    }
    if (!c)
      break;
    gdk_flush();
    usleep(5000);
  }
}

void fx_uncompress(void)
{
  int i,c;
  int py[36],dy;

  dy=1;

  if (66%speed)
    speed=11;

  if (valign)
    dy=gdk_screen_height()-65;

  for(i=0;i<bcount;i++)
    py[i]=dy+66*direc*i;

  gdk_flush();
  usleep(10000);

  while(1) {
    c=0;
    if ( bdock[bcount-1].y != py[bcount-1] )
      for(i=1;i<bcount;i++)
	if ( bdock[i].y != py[i] ) {
	  ++c;
	  bdock[i].y+=direc*speed;
	  gdk_window_move(bdock[i].square->window,bdock[i].x,bdock[i].y);
	}
    if (!c)
      break;
    gdk_flush();
    usleep(5000);
  }

}

/* uncompresses dock at startup */
gint startup(gpointer data)
{
  if (compress)
    fx_uncompress();
  compress=0;
  return FALSE;
}
